/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
import { DocumentState, interruptAndCheck, MultiMap } from "langium";
import { isFunctionDefinition, isSchemaDefinition, isSqlFile, isTableDefinition } from "./generated/ast.js";
export class SqlContainerManager {
    constructor(services) {
        this.children = new MultiMap();
        this.langiumDocuments = services.workspace.LangiumDocuments;
        services.workspace.DocumentBuilder.onBuildPhase(DocumentState.ComputedScopes, (_, cancelToken) => this.rebuild(cancelToken));
    }
    getChildren(definition) {
        return this.children.get(definition);
    }
    putChild(definition, child) {
        this.children.add(definition, child);
    }
    async rebuild(cancelToken) {
        this.children.clear();
        const documents = this.langiumDocuments.all;
        const sink = {
            schemas: [],
            tables: [],
            functions: []
        };
        for (const document of documents) {
            const value = document.parseResult.value;
            if (isSqlFile(value)) {
                await this.collectDocument(value, sink, cancelToken);
            }
        }
        await this.collectElements(sink.schemas, cancelToken);
        await this.collectElements(sink.tables, cancelToken);
        await this.collectElements(sink.functions, cancelToken);
    }
    async collectElements(elements, cancelToken) {
        var _a;
        for (const element of elements) {
            await interruptAndCheck(cancelToken);
            const reference = element.reference;
            if (reference === null || reference === void 0 ? void 0 : reference.previous) {
                const previousElement = (_a = reference.previous.element) === null || _a === void 0 ? void 0 : _a.ref;
                if (previousElement) {
                    this.putChild(previousElement, element);
                }
            }
        }
    }
    async collectDocument(file, sink, cancelToken) {
        for (const statement of file.statements) {
            await interruptAndCheck(cancelToken);
            if (isSchemaDefinition(statement)) {
                sink.schemas.push(statement);
            }
            else if (isTableDefinition(statement)) {
                sink.tables.push(statement);
            }
            else if (isFunctionDefinition(statement)) {
                sink.functions.push(statement);
            }
        }
    }
}
//# sourceMappingURL=sql-container-manager.js.map