/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
import { inject } from "langium";
import { createDefaultSharedModule, createDefaultModule } from "langium/lsp";
import { SqlServerDialectTypes } from "./dialects/sql-server/data-types.js";
import { SqlGeneratedModule, SqlGeneratedSharedModule, } from "./generated/module.js";
import { SqlCompletionProvider } from "./sql-completion-provider.js";
import { SqlContainerManager } from "./sql-container-manager.js";
import { DialectTypes } from "./sql-data-types.js";
import { SqlNameProvider } from "./sql-name-provider.js";
import { SqlScopeComputation } from "./sql-scope-computation.js";
import { SqlScopeProvider } from "./sql-scope-provider.js";
import { SqlSemanticTokenProvider } from "./sql-semantic-token-provider.js";
import { SqlTypeComputer } from "./sql-type-computation.js";
import { SqlValidationRegistry, SqlValidator } from "./sql-validator.js";
import { SqlValueConverter } from "./sql-value-converter.js";
import { SqlWorkspaceManager } from "./sql-workspace-manager.js";
export const SqlSharedModule = {
    workspace: {
        WorkspaceManager: (services) => new SqlWorkspaceManager(services),
        ContainerManager: (services) => new SqlContainerManager(services)
    }
};
/**
 * Dependency injection module that overrides Langium default services and contributes the
 * declared custom services. The Langium defaults can be partially specified to override only
 * selected services, while the custom services must be fully specified.
 */
export const SqlModule = {
    dialect: {
        dataTypes: () => new DialectTypes(SqlServerDialectTypes),
        typeComputer: services => new SqlTypeComputer(services)
    },
    parser: {
        ValueConverter: () => new SqlValueConverter(),
    },
    references: {
        NameProvider: () => new SqlNameProvider(),
        ScopeComputation: (services) => new SqlScopeComputation(services),
        ScopeProvider: (services) => new SqlScopeProvider(services),
    },
    lsp: {
        SemanticTokenProvider: (services) => new SqlSemanticTokenProvider(services),
        CompletionProvider: (services) => new SqlCompletionProvider(services),
    },
    validation: {
        ValidationRegistry: (services) => new SqlValidationRegistry(services),
        SqlValidator: (services) => new SqlValidator(services),
    },
};
/**
 * Create the full set of services required by Langium.
 *
 * First inject the shared services by merging two modules:
 *  - Langium default shared services
 *  - Services generated by langium-cli
 *
 * Then inject the language-specific services by merging three modules:
 *  - Langium default language-specific services
 *  - Services generated by langium-cli
 *  - Services specified in this file
 *
 * @param context Optional module context with the LSP connection
 * @returns An object wrapping the shared services and the language-specific services
 */
export function createSqlServices(context) {
    const shared = inject(createDefaultSharedModule(context), SqlGeneratedSharedModule, SqlSharedModule, context.sharedModule);
    const Sql = inject(createDefaultModule({ shared }), SqlGeneratedModule, SqlModule, context.module);
    shared.ServiceRegistry.register(Sql);
    return { shared, Sql };
}
//# sourceMappingURL=sql-module.js.map