/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
import { DefaultScopeProvider, assertUnreachable, stream, StreamScope, EMPTY_SCOPE, } from "langium";
import { AstUtils } from 'langium';
import { isAllTable, isColumnDefinition, isColumnNameExpression, isCommonTableExpression, isConstraintDefinition, isGlobalReference, isIndexDefinition, isKeyDefinition, isOverClause, isPrimaryKeyDefinition, isSimpleSelectStatement, isSubQuerySourceItem, isTableDefinition, isTableLike, isTableRelatedColumnExpression, isTableSourceItem, isWindowSpec, } from "./generated/ast.js";
import { getColumnCandidatesForSelectTableExpression, getColumnCandidatesForSimpleSelectStatement, getFromGlobalReference, } from "./sql-type-utilities.js";
export class SqlScopeProvider extends DefaultScopeProvider {
    constructor(services) {
        super(services);
        this.astNodeDescriptionProvider = services.workspace.AstNodeDescriptionProvider;
        this.astNodeLocator = services.workspace.AstNodeLocator;
        this.langiumDocuments = services.shared.workspace.LangiumDocuments;
        this.containerManager = services.shared.workspace.ContainerManager;
    }
    getScope(context) {
        var _a, _b, _c;
        const container = context.container;
        if (isGlobalReference(container)) {
            const previous = container.previous;
            if (previous) {
                const element = (_a = previous.element) === null || _a === void 0 ? void 0 : _a.ref;
                if (element) {
                    return this.createScopeForNodes(this.containerManager.getChildren(element));
                }
                else {
                    return EMPTY_SCOPE;
                }
            }
            else {
                // Use lexical scoping if no previous element has been assigned
                return super.getScope(context);
            }
        }
        else if (isPrimaryKeyDefinition(container)) {
            const property = context.property;
            switch (property) {
                case 'primaryKeys':
                    const tableDef = AstUtils.getContainerOfType(container, isTableDefinition);
                    return this.streamColumnDefinitions(tableDef.columns.filter(isColumnDefinition));
                default:
                    assertUnreachable(property);
            }
        }
        else if (isIndexDefinition(container)) {
            const property = context.property;
            switch (property) {
                case 'indexes':
                    const tableDef = AstUtils.getContainerOfType(container, isTableDefinition);
                    return this.streamColumnDefinitions(tableDef.columns.filter(isColumnDefinition));
                default:
                    assertUnreachable(property);
            }
        }
        else if (isKeyDefinition(container)) {
            const property = context.property;
            switch (property) {
                case 'keys':
                    const tableDef = AstUtils.getContainerOfType(container, isTableDefinition);
                    return this.streamColumnDefinitions(tableDef.columns.filter(isColumnDefinition));
                default:
                    assertUnreachable(property);
            }
        }
        else if (isConstraintDefinition(container)) {
            const property = context.property;
            switch (property) {
                case "from": {
                    const columns = AstUtils.getContainerOfType(container, isTableDefinition).columns.filter(isColumnDefinition);
                    return this.streamColumnDefinitions(columns);
                }
                case 'to': {
                    const columns = (_b = getFromGlobalReference(container.table, isTableDefinition)) === null || _b === void 0 ? void 0 : _b.columns.filter(isColumnDefinition);
                    return this.streamColumnDefinitions(columns !== null && columns !== void 0 ? columns : []);
                }
                default:
                    assertUnreachable(property);
            }
        }
        else if (isAllTable(container)) {
            const property = context.property;
            switch (property) {
                case "variableName": {
                    const selectStatement = AstUtils.getContainerOfType(container, isSimpleSelectStatement);
                    //ATTENTION! Do not recursively traverse upwards, t.* only looks up t in the current select statement
                    return this.newCaseInsensitiveScope(stream(this.getTableVariablesForSelectStatement(selectStatement)));
                }
                default:
                    assertUnreachable(property);
            }
        }
        else if (isTableRelatedColumnExpression(container)) {
            const property = context.property;
            switch (property) {
                case "variableName":
                    const selectStatement = AstUtils.getContainerOfType(container, isSimpleSelectStatement);
                    return this.getTableVariablesForSelectStatementRecursively(context, selectStatement);
                case "columnName":
                    const sourceItem = (_c = container.variableName) === null || _c === void 0 ? void 0 : _c.ref;
                    if (sourceItem) {
                        if (isTableSourceItem(sourceItem)) {
                            const tableLike = getFromGlobalReference(sourceItem.table, isTableLike);
                            if (tableLike) {
                                if (isTableDefinition(tableLike)) {
                                    return this.streamColumnDefinitions(tableLike.columns.filter(isColumnDefinition));
                                }
                                else if (isCommonTableExpression(tableLike)) {
                                    const candidates = getColumnCandidatesForSelectTableExpression(tableLike.statement);
                                    return this.streamColumnDescriptors(candidates);
                                }
                                else {
                                    assertUnreachable(tableLike);
                                }
                            }
                        }
                        else if (isSubQuerySourceItem(sourceItem)) {
                            const selectStatement = sourceItem.subQuery;
                            const candidates = getColumnCandidatesForSelectTableExpression(selectStatement);
                            return this.streamColumnDescriptors(candidates);
                        }
                        else {
                            assertUnreachable(sourceItem);
                        }
                    }
                    break;
                default:
                    assertUnreachable(property);
            }
        }
        else if (isColumnNameExpression(container)) {
            const property = context.property;
            switch (property) {
                case 'columnName':
                    const selectStatement = AstUtils.getContainerOfType(container, isSimpleSelectStatement);
                    const candidates = getColumnCandidatesForSimpleSelectStatement(selectStatement);
                    return this.streamColumnDescriptors(candidates);
                default:
                    assertUnreachable(property);
            }
        }
        else if (isOverClause(container)) {
            const property = context.property;
            switch (property) {
                case 'windowName':
                    const selectStatement = AstUtils.getContainerOfType(container, isSimpleSelectStatement);
                    const nodes = AstUtils.streamAllContents(selectStatement).filter(isWindowSpec).toArray();
                    return this.packToDescriptions(nodes);
            }
        }
        else {
            assertUnreachable(container);
        }
        return super.getScope(context);
    }
    getTableVariablesForSelectStatementRecursively(context, selectStatement) {
        let outerScope = undefined;
        if (selectStatement && AstUtils.hasContainerOfType(selectStatement.$container, isSimpleSelectStatement)) {
            const outerSelectStatement = AstUtils.getContainerOfType(selectStatement.$container, isSimpleSelectStatement);
            outerScope = this.getTableVariablesForSelectStatementRecursively(context, outerSelectStatement);
        }
        const descriptions = this.getTableVariablesForSelectStatement(selectStatement);
        return this.newCaseInsensitiveScope(stream(descriptions), outerScope);
    }
    getTableVariablesForSelectStatement(selectStatement) {
        if (selectStatement === null || selectStatement === void 0 ? void 0 : selectStatement.from) {
            const astDescriptions = [];
            for (const source of selectStatement.from.sources.list) {
                const items = [source.item].concat(source.joins.map((j) => j.nextItem));
                for (const item of items) {
                    if (item.name) {
                        astDescriptions.push(this.astNodeDescriptionProvider.createDescription(item, item.name));
                    }
                    else {
                        if (isTableSourceItem(item)) {
                            const tableLike = getFromGlobalReference(item.table, isTableLike);
                            if (tableLike) {
                                const name = this.nameProvider.getName(tableLike);
                                if (name) {
                                    astDescriptions.push(this.astNodeDescriptionProvider.createDescription(item, name));
                                }
                            }
                        }
                        else if (!isSubQuerySourceItem(item)) {
                            assertUnreachable(item);
                        }
                    }
                }
            }
            return astDescriptions;
        }
        return [];
    }
    getGlobalScope(nodeType, _context) {
        return this.newCaseInsensitiveScope(this.indexManager.allElements(nodeType));
    }
    packToDescriptions(nodes) {
        return this.newCaseInsensitiveScope(stream(nodes.map((c) => this.astNodeDescriptionProvider.createDescription(c, c.name))));
    }
    streamColumnDescriptors(columns) {
        return this.newCaseInsensitiveScope(stream(columns.filter(c => c.name).map(c => this.astNodeDescriptionProvider.createDescription(c.node, c.name))));
    }
    streamColumnDefinitions(columns) {
        return this.packToDescriptions(columns);
    }
    newCaseInsensitiveScope(stream, outerScope = undefined) {
        return new StreamScope(stream, outerScope, { caseInsensitive: true });
    }
}
//# sourceMappingURL=sql-scope-provider.js.map