/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
import { assertUnreachable } from "langium";
import * as AST from "./generated/ast.js";
export function getFromGlobalReference(reference, is) {
    var _a;
    const element = (_a = reference === null || reference === void 0 ? void 0 : reference.element) === null || _a === void 0 ? void 0 : _a.ref;
    if (element && is(element)) {
        return element;
    }
    else {
        return undefined;
    }
}
export function getDefinitionType(element) {
    if (AST.isCatalogDefinition(element)) {
        return DefinitionType.Catalog;
    }
    else if (AST.isSchemaDefinition(element)) {
        return DefinitionType.Schema;
    }
    else {
        return DefinitionType.Table;
    }
}
export function getDefinitionTypeName(element) {
    if (AST.isCatalogDefinition(element)) {
        return 'DATABASE';
    }
    else if (AST.isSchemaDefinition(element)) {
        return 'SCHEMA';
    }
    else if (AST.isTableLike(element)) {
        return 'TABLE';
    }
    else if (AST.isFunctionDefinition(element)) {
        return 'FUNCTION';
    }
    else {
        return 'UNKNOWN';
    }
}
export var DefinitionType;
(function (DefinitionType) {
    DefinitionType[DefinitionType["Catalog"] = 0] = "Catalog";
    DefinitionType[DefinitionType["Schema"] = 1] = "Schema";
    DefinitionType[DefinitionType["Table"] = 2] = "Table";
})(DefinitionType = DefinitionType || (DefinitionType = {}));
export function getColumnsForSelectTableExpression(selectTableExpression, onlyAliases = false) {
    if (AST.isBinaryTableExpression(selectTableExpression)) {
        const lhs = getColumnsForSelectTableExpression(selectTableExpression.left, onlyAliases);
        //const rhs = getColumnsForSelectTableExpression(selectTableExpression.right, onlyAliases);
        return lhs;
    }
    else if (AST.isParenthesesSelectTableExpression(selectTableExpression)) {
        return getColumnsForSelectTableExpression(selectTableExpression.value);
    }
    else if (AST.isSimpleSelectTableExpression(selectTableExpression)) {
        return getColumnsSimpleSelectStatement(selectTableExpression.select, onlyAliases);
    }
    else if (!selectTableExpression) {
        return [];
    }
    else {
        assertUnreachable(selectTableExpression);
    }
}
function getColumnsSimpleSelectStatement(simpleSelectStatement, onlyAliases) {
    if (!simpleSelectStatement || !simpleSelectStatement.select) {
        return [];
    }
    return simpleSelectStatement.select.elements.flatMap(e => {
        var _a;
        if (AST.isAllStar(e)) {
            if (onlyAliases) {
                return [];
            }
            const fromAllSources = getColumnCandidatesForSimpleSelectStatement(simpleSelectStatement);
            return fromAllSources.flatMap(t => t);
        }
        else if (AST.isAllTable(e)) {
            if (onlyAliases) {
                return [];
            }
            if (!simpleSelectStatement.from) {
                return [];
            }
            const ref = e.variableName.ref;
            if (AST.isTableSourceItem(ref)) {
                const tableLike = getFromGlobalReference(ref.table, AST.isTableLike);
                if (AST.isTableDefinition(tableLike)) {
                    const columns = (_a = tableLike.columns.filter(AST.isColumnDefinition)) !== null && _a !== void 0 ? _a : [];
                    return columns.map(c => ({
                        name: c.name,
                        typedNode: c.dataType,
                        node: c,
                        isScopedByVariable: true
                    }));
                }
                else if (AST.isCommonTableExpression(tableLike)) {
                    const columns = getColumnsForSelectTableExpression(tableLike.statement);
                    if (tableLike.columnNames.length > 0) {
                        return columns.map((c, i) => (Object.assign(Object.assign({}, c), { name: tableLike.columnNames[i].name })));
                    }
                    return columns;
                }
                else if (tableLike) {
                    assertUnreachable(tableLike);
                }
            }
            else if (AST.isSubQuerySourceItem(ref)) {
                return getColumnsForSelectTableExpression(ref.subQuery);
            }
            else {
                assertUnreachable(ref);
            }
        }
        else if (AST.isExpressionQuery(e)) {
            if (e.name) {
                return [{
                        name: e.name,
                        typedNode: e.expr,
                        node: e,
                        isScopedByVariable: false
                    }];
            }
            else {
                if (onlyAliases) {
                    return [];
                }
                const expr = e.expr;
                if (AST.isTableRelatedColumnExpression(expr)) {
                    return resolveColumnNameTypedNode(e, expr.columnName);
                }
                else if (AST.isFunctionCall(expr)) {
                    const functionLike = getFromGlobalReference(expr.function, AST.isFunctionDefinition);
                    if (functionLike) {
                        const name = expr.function.element.$refText;
                        return [{
                                name,
                                isScopedByVariable: false,
                                node: e,
                                typedNode: functionLike.returnType
                            }];
                    }
                    else {
                        return [];
                    }
                }
                else if (AST.isColumnNameExpression(expr)) {
                    const fromAllSources = getColumnCandidatesForSimpleSelectStatement(simpleSelectStatement);
                    const name = expr.columnName.$refText;
                    const column = fromAllSources.find(s => !s.isScopedByVariable && s.name === name);
                    if (column) {
                        return [column];
                    }
                    else {
                        return resolveColumnNameTypedNode(expr, expr.columnName);
                    }
                }
                else if (AST.isSubQueryExpression(expr)) {
                    const columns = getColumnsForSelectTableExpression(expr.subQuery);
                    return [columns[0]];
                }
                else {
                    return [{
                            node: e,
                            typedNode: e.expr,
                            isScopedByVariable: false
                        }];
                }
            }
        }
        else {
            assertUnreachable(e);
        }
        return [];
    });
}
function resolveColumnNameTypedNode(expression, columnName) {
    const ref = columnName.ref;
    let typed = expression;
    if (AST.isExpressionQuery(ref)) {
        typed = ref.expr;
    }
    else if (AST.isColumnDefinition(ref)) {
        typed = ref.dataType;
    }
    return [{
            name: ref === null || ref === void 0 ? void 0 : ref.name,
            isScopedByVariable: false,
            node: expression,
            typedNode: typed
        }];
}
export function getColumnCandidatesForSelectTableExpression(selectTableExpression) {
    if (AST.isBinaryTableExpression(selectTableExpression)) {
        const lhs = getColumnCandidatesForSelectTableExpression(selectTableExpression.left);
        //const rhs = getColumnCandidatesForSelectTableExpression(selectTableExpression.right);
        return lhs;
    }
    else if (AST.isParenthesesSelectTableExpression(selectTableExpression)) {
        return getColumnCandidatesForSelectTableExpression(selectTableExpression.value);
    }
    else if (AST.isSimpleSelectTableExpression(selectTableExpression)) {
        return getColumnCandidatesForSimpleSelectStatement(selectTableExpression.select);
    }
    else if (!selectTableExpression) {
        return [];
    }
    {
        assertUnreachable(selectTableExpression);
    }
}
export function getColumnCandidatesForSimpleSelectStatement(selectStatement) {
    var _a, _b;
    const selectElementColumns = getColumnsSimpleSelectStatement(selectStatement, true);
    const fromComputedColumns = (_b = (_a = selectStatement === null || selectStatement === void 0 ? void 0 : selectStatement.from) === null || _a === void 0 ? void 0 : _a.sources.list.flatMap(getColumnsForTableSource)) !== null && _b !== void 0 ? _b : [];
    return selectElementColumns.concat(fromComputedColumns);
}
function getColumnsForTableSource(source) {
    const items = [source.item].concat(source.joins.map(j => j.nextItem));
    return items.flatMap(item => {
        if (AST.isTableSourceItem(item)) {
            const tableLike = getFromGlobalReference(item.table, AST.isTableLike);
            if (AST.isTableDefinition(tableLike)) {
                return tableLike.columns.filter(AST.isColumnDefinition).map(column => ({
                    name: column.name,
                    typedNode: column.dataType,
                    node: column,
                    isScopedByVariable: item.name != null
                }));
            }
            else if (AST.isCommonTableExpression(tableLike)) {
                let columns = getColumnsForSelectTableExpression(tableLike.statement);
                if ((tableLike === null || tableLike === void 0 ? void 0 : tableLike.columnNames.length) > 0) {
                    columns = columns.map((c, i) => { var _a; return (Object.assign(Object.assign({}, c), { name: (_a = tableLike.columnNames[i]) === null || _a === void 0 ? void 0 : _a.name })); });
                }
                return columns;
            }
            return [];
        }
        else if (AST.isSubQuerySourceItem(item)) {
            return getColumnsForSelectTableExpression(item.subQuery);
        }
        else {
            assertUnreachable(item);
        }
    });
}
//# sourceMappingURL=sql-type-utilities.js.map