/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
import { ValidationRegistry, } from "langium";
import * as ast from "./generated/ast.js";
import _ from "lodash";
import { ReportAs } from "./sql-error-codes.js";
import { isTypeABoolean } from "./sql-type-descriptors.js";
import { getColumnsForSelectTableExpression, getDefinitionType, getDefinitionTypeName } from "./sql-type-utilities.js";
import { canConvert } from "./sql-type-conversion.js";
import { isCompatibleWithDefinition } from "./sql-data-types.js";
export class SqlValidationRegistry extends ValidationRegistry {
    constructor(services) {
        super(services);
        const validator = services.validation.SqlValidator;
        const checks = {
            TableDefinition: [validator.checkIfTableDefinitionHasAtLeastOneColumn],
            SimpleSelectStatement: [
                validator.checkVariableNamesAreUnique,
                validator.checkIfSelectStatementWithAllStarSelectItemHasAtLeastOneTableSource
            ],
            //Expression: [validator.checkIfExpressionHasType], //TODO uncomment when type system is bullet-proof
            BinaryTableExpression: [validator.checkBinaryTableExpressionMatches],
            NumberLiteral: [validator.checkIntegerLiteralIsWholeNumber],
            BinaryExpression: [validator.checkBinaryExpressionType],
            UnaryExpression: [validator.checkUnaryExpressionType],
            WhereClause: [validator.checkWhereIsBoolean],
            HavingClause: [validator.checkHavingIsBoolean],
            SubQueryExpression: [validator.checkIfSubQuerySelectsExactlyOneValue],
            ReferenceDefinition: [validator.checkIfReferencePointsToCorrectParent],
            FunctionCall: [validator.checkFunctionCallTarget],
            TableSourceItem: [validator.checkTableSourceItemTarget],
            ConstraintDefinition: [validator.checkConstraintTarget],
            DataType: [validator.checkIfKnownDataType],
        };
        this.register(checks, validator);
    }
}
export class SqlValidator {
    constructor(services) {
        this.typeComputer = services.dialect.typeComputer;
        this.dataTypes = services.dialect.dataTypes.allTypes();
    }
    checkBinaryTableExpressionMatches(expr, accept) {
        const lhs = getColumnsForSelectTableExpression(expr.left);
        const rhs = getColumnsForSelectTableExpression(expr.right);
        if (lhs.length !== rhs.length) {
            ReportAs.TableOperationUsesTablesWithDifferentColumnCounts(expr, {}, accept);
        }
        else {
            lhs.forEach((left, index) => {
                const right = rhs[index];
                const leftType = this.typeComputer.computeType(left.typedNode);
                const rightType = this.typeComputer.computeType(right.typedNode);
                if (leftType && rightType) {
                    if (!canConvert(leftType, rightType) && !canConvert(rightType, leftType)) {
                        ReportAs.TableOperationUsesTablesWithDifferentColumnTypes(expr, { columnIndex: index }, accept);
                    }
                } //else should be handled by different validator
            });
        }
    }
    checkIfExpressionHasType(expr, accept) {
        const type = this.typeComputer.computeType(expr);
        if (!type) {
            ReportAs.CannotDeriveTypeOfExpression(expr, {}, accept);
        }
    }
    checkWhereIsBoolean(clause, accept) {
        const type = this.typeComputer.computeType(clause.rowCondition);
        if (type && !isTypeABoolean(type)) {
            ReportAs.ExpressionMustReturnABoolean(clause.rowCondition, type, accept);
        }
    }
    checkHavingIsBoolean(clause, accept) {
        const type = this.typeComputer.computeType(clause.groupCondition);
        if (type && !isTypeABoolean(type)) {
            ReportAs.ExpressionMustReturnABoolean(clause.groupCondition, type, accept);
        }
    }
    checkBinaryExpressionType(expr, accept) {
        const left = this.typeComputer.computeType(expr.left);
        const right = this.typeComputer.computeType(expr.right);
        const returnType = this.typeComputer.computeType(expr);
        if (left && right && !returnType) {
            ReportAs.BinaryOperatorNotDefinedForGivenExpressions(expr, {
                left,
                right,
                op: expr.operator
            }, accept);
        }
    }
    checkUnaryExpressionType(expr, accept) {
        const operand = this.typeComputer.computeType(expr.value);
        const returnType = this.typeComputer.computeType(expr);
        if (operand && !returnType) {
            ReportAs.UnaryOperatorNotDefinedForGivenExpression(expr, {
                operand,
                op: expr.operator
            }, accept);
        }
    }
    checkVariableNamesAreUnique(query, accept) {
        var _a, _b;
        const groups = _.groupBy((_b = (_a = query.from) === null || _a === void 0 ? void 0 : _a.sources) === null || _b === void 0 ? void 0 : _b.list, (s) => { var _a; return (_a = s.item) === null || _a === void 0 ? void 0 : _a.name; });
        for (const [key, group] of Object.entries(groups).filter((g) => g[0] && g[1].length > 1)) {
            for (const member of group.filter(e => e.item)) {
                ReportAs.DuplicatedVariableName(member.item, { name: key }, accept);
            }
        }
    }
    checkIntegerLiteralIsWholeNumber(literal, accept) {
        // if (Math.floor(literal.value) !== literal.value) {
        //     ReportAs.NumericValueIsNotInteger(
        //         literal,
        //         { value: literal.value },
        //         accept
        //     );
        // }
    }
    checkIfSelectStatementWithAllStarSelectItemHasAtLeastOneTableSource(selectStatement, accept) {
        var _a;
        if (((_a = selectStatement.select) === null || _a === void 0 ? void 0 : _a.elements.filter(ast.isAllStar).length) > 0) {
            if (!selectStatement.from) {
                ReportAs.AllStarSelectionRequiresTableSources(selectStatement, {}, accept);
            }
        }
    }
    checkIfTableDefinitionHasAtLeastOneColumn(tableDefinition, accept) {
        if (tableDefinition.columns.length === 0) {
            ReportAs.TableDefinitionRequiresAtLeastOneColumn(tableDefinition.reference, {}, accept);
        }
    }
    checkIfReferencePointsToCorrectParent(referenceDefinition, accept) {
        this.checkReferenceParent(referenceDefinition.reference, accept);
    }
    checkReferenceParent(globalReference, accept) {
        var _a, _b, _c;
        const element = (_a = globalReference === null || globalReference === void 0 ? void 0 : globalReference.element) === null || _a === void 0 ? void 0 : _a.ref;
        if (element) {
            const parent = (_c = (_b = globalReference.previous) === null || _b === void 0 ? void 0 : _b.element) === null || _c === void 0 ? void 0 : _c.ref;
            if (parent) {
                const ownType = getDefinitionType(element);
                const parentType = getDefinitionType(parent);
                if (parentType >= ownType) {
                    accept('error', `Cannot nest element of type '${getDefinitionTypeName(element)}' inside of '${getDefinitionTypeName(parent)}'`, {
                        node: globalReference,
                        property: 'previous'
                    });
                }
                else if (globalReference.previous) {
                    this.checkReferenceParent(globalReference.previous, accept);
                }
            }
        }
    }
    //TODO does not hold for insertions! INSERT INTO employees SELECT id, name FROM xxx
    checkIfSubQuerySelectsExactlyOneValue(subQueryExpression, accept) {
        const type = this.typeComputer.computeTypeOfSelectStatement(subQueryExpression.subQuery);
        if (type.discriminator === 'row' && type.columnTypes.length > 1) {
            ReportAs.SubQueriesWithinSelectStatementsMustHaveExactlyOneColumn(subQueryExpression, {}, accept);
        }
    }
    checkFunctionCallTarget(functionCall, accept) {
        var _a, _b;
        const reference = (_b = (_a = functionCall.function) === null || _a === void 0 ? void 0 : _a.element) === null || _b === void 0 ? void 0 : _b.ref;
        if (reference && !ast.isFunctionDefinition(reference)) {
            ReportAs.IncorrectGlobalReferenceTarget(functionCall.function, {
                expected: 'FUNCTION',
                received: getDefinitionTypeName(reference)
            }, accept);
        }
    }
    checkTableSourceItemTarget(tableSourceItem, accept) {
        var _a, _b;
        const reference = (_b = (_a = tableSourceItem.table) === null || _a === void 0 ? void 0 : _a.element) === null || _b === void 0 ? void 0 : _b.ref;
        if (reference && !ast.isTableLike(reference)) {
            ReportAs.IncorrectGlobalReferenceTarget(tableSourceItem.table, {
                expected: 'TABLE',
                received: getDefinitionTypeName(reference)
            }, accept);
        }
    }
    checkConstraintTarget(constraintDefinition, accept) {
        var _a, _b;
        const reference = (_b = (_a = constraintDefinition.table) === null || _a === void 0 ? void 0 : _a.element) === null || _b === void 0 ? void 0 : _b.ref;
        if (reference && !ast.isTableDefinition(reference)) {
            ReportAs.IncorrectGlobalReferenceTarget(constraintDefinition.table, {
                expected: 'TABLE',
                received: getDefinitionTypeName(reference)
            }, accept);
        }
    }
    checkIfKnownDataType(dataType, accept) {
        if (!this.dataTypes.some(dt => isCompatibleWithDefinition(dataType, dt))) {
            ReportAs.UnknownDataType(dataType, { dataType }, accept);
        }
    }
}
//# sourceMappingURL=sql-validator.js.map